// The contents of this file are used with permission, subject to
// the Mozilla Public License Version 1.1 (the "License"); you may
// not use this file except in compliance with the License. You may
// obtain a copy of the License at
// http://www.mozilla.org/MPL/MPL-1.1.html
//
// Software distributed under the License is distributed on an
// "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
unit DnHttpParser;
interface
uses  SysUtils, Windows, Classes, contnrs,
      DnRtl, DnConst;

type

  TDnHttpParser = class(TObject)
  protected
    FHttpData: String;
    FHttpVersion: String;
    FMethodName: String;
    FMethodURL: String;
    FHttpResponseCode: Integer;
    FHttpResponseReason: String;
    FNames: TStringList;
    FValues: TStringList;
    
    class function  ExtractLine(var Data: String): String;
    class procedure ExtractPair(var Data: String; var LeftPart, RightPart: String);
    class procedure StripSpaces(var Data: String);
    class function  Skip1Space(const S: String; StartWith: Integer): Integer;
    class function  Skip1Colon(const S: String; StartWith: Integer): Integer;
    class function  GetWord(const S: String; StartWith: Integer; var Lexem: String): Integer;
    class function  GetIntLexem(const S: String; StartWith: Integer; var Lexem: String): Integer;
    procedure ParseHttpHeaders(S: String);
    function  AssembleHttpHeaders: String;

    function  GetHttpHeader(const Name: String): String;
    procedure SetHttpHeader(const Name, Value: String);

  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Clear;
    procedure   ParseResponse(const HttpData: String);
    procedure   ParseRequest(const HttpData: String);
    function    AssembleResponse: String;
    function    AssembleRequest: String;

    class function  IsAbsoluteUrl(const Url: String): Boolean;

    class procedure ParseAbsoluteUrl(const Url: String;
        var Protocol, User, Password, Host, Port, Path,
        Query: String; var UserExists, PasswordExists: Boolean); overload;

    class procedure ParseRelativeUrl(const Url: String; var Path, Query: String);
    class procedure ParseHttpTime(const S: String; StartWith: Integer;
        var FinishWith: Integer;
        var Year, Month, Day, DayOfWeek, Hour, Minute, Second: Integer;
        var TimeZone: String);
    class function  FormatHttpTime(Year, Month, Day, DayOfWeek, Hour,
        Minute, Second: Integer;
        TimeZone: String): String;
    class procedure ParseResponseContentType(const S: String; var ContentType: String;
        var CharSet: String);
    class function  ParseAcceptList(const S: String): TStringList;
    class function  AdjustSlash(const S: String): String;
    property  HttpMethodName: String read FMethodName write FMethodName;
    property  HttpMethodURL: String read FMethodURL write FMethodURL;
    property  HttpVersion: String read FHttpVersion write FHttpVersion;
    property  HttpHeader[const Name: String]: String read GetHttpHeader write SetHttpHeader;
    property  HttpCode: Integer read FHttpResponseCode write FHttpResponseCode;
    property  HttpReason: String read FHttpResponseReason write FHttpResponseReason;
  end;

  
implementation
const
  CRLF: String = #13#10;
  ShortWeekDays: array[0..6] of String =
    ( 'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
  LongWeekDays: array[0..6] of String =
    ( 'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday' );
    
  MonthNames: array[1..12] of String = ('Jan', 'Feb', 'Mar', 'Apr',
                                        'May', 'Jun', 'Jul', 'Aug',
                                        'Sep', 'Oct', 'Nov', 'Dec');


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

class function  TDnHttpParser.Skip1Space(const S: String; StartWith: Integer): Integer;
var i: Integer;
begin
  i := StartWith;
  if (i < Length(S)) and (S[i] = ' ') then
    Result := i + 1
  else
    raise EDnException.Create(ErrCannotParseHttpHeader, 0, S);
end;

class function  TDnHttpParser.Skip1Colon(const S: String; StartWith: Integer): Integer;
var i: Integer;
begin
  i := StartWith;
  if (i < Length(S)) and (S[i] = ':') then
    Result := i + 1
  else
    raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
end;

class function  TDnHttpParser.GetWord(const S: String; StartWith: Integer; var Lexem: String): Integer;
var i, len: Integer;
begin
  i := StartWith; len := Length(S);
  while (i<=len) and (S[i] in ['A'..'Z','a'..'z']) do
    Inc(i);
  Lexem := Copy(S, StartWith, i-StartWith);
  Result := i;
end;

class function  TDnHttpParser.GetIntLexem(const S: String; StartWith: Integer; var Lexem: String): Integer;
var i, len: Integer;
begin
  i := StartWith; len := Length(S);
  while (i<=len) and ((S[i] in ['0'..'9']) or ((S[i] in ['-', '+']) and (i = StartWith))) do
    Inc(i);
  Lexem := Copy(S, StartWith, i-StartWith);
  Result := i;
end;

class procedure TDnHttpParser.ParseResponseContentType(const S: String; var ContentType: String;
                                                      var CharSet: String);
var semicolon: Integer;
begin
  semicolon := Pos(';', S);
  if semicolon <> 0 then
  begin
    ContentType := Copy(S, 1, semicolon-1);
    CharSet := Copy(S, semicolon+1, Length(S) - semicolon);
  end else
  begin
    ContentType := Copy(S, 1, Length(S));
    CharSet := 'us-ascii';
  end;
  ContentType := Trim(ContentType); CharSet := Trim(CharSet);
end;

class function  TDnHttpParser.ParseAcceptList(const S: String): TStringList;
begin
  Result := TStringList.Create;
end;

class function  TDnHttpParser.AdjustSlash(const S: String): String;
var i: Integer;
begin
  Result := S;
  for i:=1 to Length(S) do
    if Result[i] = '/' then
      Result[i] := '\'; 
end;

class procedure TDnHttpParser.ParseHttpTime(const S: String; StartWith: Integer;
                                            var FinishWith: Integer;
                                            var Year, Month, Day, DayOfWeek, Hour,
                                            Minute, Second: Integer;
                                            var TimeZone: String);
var i, len, j, lexemLen, TimeKind: Integer;
    Lexem: String;
begin
  i := StartWith; len := Length(S);
  i := GetWord(S, i, Lexem); lexemLen := i - StartWith;

  if lexemLen > 3 then
    TimeKind := 2 //RFC 850
  else if (i <= len) and (S[i] = ',') and (lexemLen = 3) then
    TimeKind := 1 //RFC 822
  else if (i <= len) and (S[i] = ' ') and (lexemLen = 3) then
    TimeKind := 3 //asctime
  else if lexemLen = 0 then
    TimeKind := 4 //number of seconds
  else
    raise EDnException.Create(ErrCannotParseHttpTime, 0, S);

  case TimeKind of
    1:  begin //RFC 822
          //find a number of week day
          j := 0;
          while (j < 7) and (ShortWeekDays[j] <> Lexem) do
            Inc(j);
          if j = 7 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
          DayOfWeek := j;

          Inc(i); //skip a comma
          //find a number of day (in month)
          i := Skip1Space(S, i); i := GetIntLexem(S, i, Lexem);
          Day := StrToInt(Lexem);
          if Day > 31 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);

          i := Skip1Space(S, i);
          i := GetWord(S, i, Lexem);
          j := 1;
          while (j<13) and (MonthNames[j] <> Lexem) do
            Inc(j);
          if j = 13 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
          Month := j;

          //find a year
          i := Skip1Space(S, i);
          i := GetIntLexem(S, i, Lexem);
          Year := StrToInt(Lexem);

          //find a hour
          i := Skip1Space(S, i); i := GetIntLexem(S, i, Lexem);
          Hour := StrToInt(Lexem);

          //find a minute
          i := Skip1Colon(S, i); i := GetIntLexem(S, i, Lexem);
          Minute := StrToInt(Lexem);

          //find a second
          i := Skip1Colon(S, i); i := GetIntLexem(S, i, Lexem);
          Second := StrToInt(Lexem);

          i := Skip1Space(S, i); i := GetWord(S, i, Lexem);
          TimeZone := Lexem;
          FinishWith := i-1;
        end;
    2:  begin //RFC 850
          //find a number of week day
          j := 0;
          while (j < 7) and (LongWeekDays[j] <> Lexem) do
            Inc(j);
          if j = 7 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
          DayOfWeek := j;

          Inc(i); //skip a comma
          //find a number of day (in month)
          i := Skip1Space(S, i); i := GetIntLexem(S, i, Lexem);
          Day := StrToInt(Lexem);
          if Day > 31 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);

          Inc(i);//skip a '-'
          i := GetWord(S, i, Lexem);
          j := 1;
          while (j<13) and (MonthNames[j] <> Lexem) do
            Inc(j);
          if j = 13 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
          Month := j;
          Inc(i);//skip a '-'
          i := GetIntLexem(S, i, Lexem);
          Year := StrToInt(Lexem);
          //Y2K problem
          if Year > 35 then
            Year := Year + 1900
          else
            Year := Year + 2000;

          //find a hour
          i := Skip1Space(S, i); i := GetIntLexem(S, i, Lexem);
          Hour := StrToInt(Lexem);

          //find a minute
          i := Skip1Colon(S, i); i := GetIntLexem(S, i, Lexem);
          Minute := StrToInt(Lexem);

          //find a second
          i := Skip1Colon(S, i); i := GetIntLexem(S, i, Lexem);
          Second := StrToInt(Lexem);

          i := Skip1Space(S, i); i := GetWord(S, i, Lexem);
          TimeZone := Lexem;
          FinishWith := i-1;
        end;
    3:  begin
          //find a number of week day
          j := 0;
          while (j < 7) and (ShortWeekDays[j] <> Lexem) do
            Inc(j);
          if j = 7 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
          DayOfWeek := j;
          i := Skip1Space(S, i); i := GetWord(S, i, Lexem);
          j := 1;
          while (j<13) and (MonthNames[j] <> Lexem) do
            Inc(j);
          if j = 13 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
          Month := j;
          i := Skip1Space(S, i);
          if i>Length(S) then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);
          if S[i] = ' ' then
            i := Skip1Space(S, i);

          i := GetIntLexem(S, i, Lexem);
          day := StrToInt(Lexem);
          if Day > 31 then
            raise EDnException.Create(ErrCannotParseHttpTime, 0, S);

          //find a hour
          i := Skip1Space(S, i); i := GetIntLexem(S, i, Lexem);
          Hour := StrToInt(Lexem);

          //find a minute
          i := Skip1Colon(S, i); i := GetIntLexem(S, i, Lexem);
          Minute := StrToInt(Lexem);

          //find a second
          i := Skip1Colon(S, i); i := GetIntLexem(S, i, Lexem);
          Second := StrToInt(Lexem);

          i := Skip1Space(S, i); i := GetIntLexem(S, i, Lexem);
          Year := StrToInt(Lexem);
          TimeZone := 'GMT';
          FinishWith := i-1;
        end;
    4:  begin
          i := GetIntLexem(S, i, Lexem);
          Second := StrToInt(Lexem);
          FinishWith := i-1;
        end;
  end;

end;

class function  TDnHttpParser.FormatHttpTime( Year, Month, Day, DayOfWeek, Hour, Minute, Second: Integer;
                                              TimeZone: String): String;
var HttpTime: String;
begin
  if  (DayOfWeek > 6) or (DayOfWeek < 0) or (Month > 12) or (Month < 1) or
      (Day < 1) or (Day > 31) or (Hour > 24) or (Hour < 0) or
      (Minute < 0) or (Minute > 59) or (Second < 0) or (Second > 59) then
    raise EDnException.Create(ErrInvalidParameter, -1);
    
  HttpTime := '%s, %.2d %s %.4d %.2d:%.2:%.2 %s';
  Result := Format (HttpTime, [ShortWeekDays[DayOfWeek], Day,
                    MonthNames[Month], Year, Hour,
                    Minute, Second, TimeZone]);
end;


constructor TDnHttpParser.Create;
begin
  inherited Create;
  FHttpData := '';
  FNames := TStringList.Create;
  FValues := TStringList.Create;
end;

procedure TDnHttpParser.Clear;
begin
  FNames.Clear;
  FValues.Clear;
  FHttpVersion := '';
  FMethodName := '';
  FMethodURL := '';
  FHttpResponseCode := 0;
  FHttpResponseReason := '';
end;


destructor TDnHttpParser.Destroy;
begin
  FreeAndNil(FNames);
  FreeAndNil(FValues);
  inherited Destroy;
end;

function  TDnHttpParser.GetHttpHeader(const Name: String): String;
var i: Integer;
begin
  i := FNames.IndexOf(Name);
  if i = -1 then
    Result := ''
  else
    Result := FValues[i];
end;

procedure TDnHttpParser.SetHttpHeader(const Name, Value: String);
var i: Integer;
begin
  i := FNames.IndexOf(Name);
  if i <> -1 then
    FValues[i] := Value
  else
  begin
    FNames.Add(Name);
    FValues.Add(Value);
  end;
end;

procedure TDnHttpParser.ParseHttpHeaders(S: String);
var EOL, AssignPos: Integer;
    Line: String;
begin
  EOL := Pos(CRLF, S);
  while EOL<>0 do
  begin
    Line := Copy(S, 1, EOL-1); Delete(S, 1, EOL+1);
    AssignPos := Pos('=', S);
    if AssignPos <> 0 then
    begin
      FNames.Add(Trim(Copy(Line, 1, AssignPos-1)));
      FValues.Add(Trim(Copy(Line, AssignPos+1, Length(Line) - AssignPos)));
    end;
    EOL := Pos(CRLF, S);
  end;
end;

procedure TDnHttpParser.ParseRequest(const HttpData: String);
var S: String;
    Line, Lexem: String;
    EOL, i, j: Integer;
begin
  FNames.Clear; FValues.Clear;
  //Normalize CRLF pairs
  S := AdjustLineBreaks(HttpData);
  EOL := Pos(CRLF, S);
  Line := Copy(S, 1, EOL-1);
  Delete(S, 1, EOL+1);

  //parse method name, URL, HTTP version
  i := GetWord(Line, 1, Lexem);
  FMethodName := UpperCase(Lexem);
  i := Skip1Space(Line, i);
  j := i;
  while (Line[j] <> ' ') and (j < Length(Line)) do
    Inc(j);
  FMethodURL := Copy(Line, i, j-i);
  i := j;
  FHttpVersion := UpperCase(Copy(Line, i, Length(Line) - i + 1));
  ParseHttpHeaders(S);
end;

procedure TDnHttpParser.ParseResponse(const HttpData: String);
var S, Line: String;
    SpacePos, EOL: Integer;
begin
  FNames.Clear; FValues.Clear;
  S := AdjustLineBreaks(HttpData);
  EOL := Pos (CRLF, S);
  if EOL = 0 then
    raise EDnException.Create(ErrCannotParseHttp, 0, S);
  Line := Copy(S, 1, EOL-1); Delete(S, 1, EOL+1);
  SpacePos := Pos(' ', S);
  if SpacePos = 0 then
    raise EDnException.Create(ErrCannotParseHttp, 0, Line);
  FHttpVersion := Copy(Line, 1, SpacePos-1); Delete(Line, 1, SpacePos);
  SpacePos := Pos(' ', Line);
  if SpacePos = 0 then
    raise EDnException.Create(ErrCannotParseHttp, 0, Line);
  FHttpResponseCode := StrToInt(Copy(Line, 1, SpacePos-1));
  Delete(Line, 1, SpacePos+1);
  FHttpResponseReason := Copy(Line, 1, Length(Line));
  ParseHttpHeaders(S);
end;

function TDnHttpParser.AssembleHttpHeaders: String;
var i: Integer;
begin
  Result := '';
  for i:=0 to FNames.Count-1 do
    Result:= Result + FNames[i] + ': ' + FValues[i] + CRLF; 
end;

function TDnHttpParser.AssembleResponse: String;
begin
  Result := FHttpVersion + ' ' + IntToStr(FHttpResponseCode) + ' ' + FHttpResponseReason + CRLF;
  Result := Result + AssembleHttpHeaders();
  Result := Result + CRLF;
end;

function TDnHttpParser.AssembleRequest: String;
begin
  Result := FMethodName + ' ' + FMethodURL + ' ' + FHttpVersion + CRLF;
  Result := Result + AssembleHttpHeaders();
  Result := Result + CRLF;
end;

class function TDnHttpParser.ExtractLine(var Data: String): String;
var CrLfPos: Integer;
begin
  CrLfPos := Pos(CRLF, Data);
  if CrLfPos > 0 then
  begin
    Result := Copy(Data, 1, CrLfPos-1);
    Delete(Data, 1, CrLfPos+1);
  end else
  begin
    Result := Data;
    Data := '';
  end;
end;


class procedure TDnHttpParser.ExtractPair(var Data: String; var LeftPart, RightPart: String);
var ScPos: Integer;
begin
  ScPos := Pos(':', Data);
  if ScPos > 0 then
  begin
    LeftPart := Copy(Data, 1, ScPos-1);
    RightPart := Copy(Data, ScPos+1, Length(Data) - ScPos);
  end else
    LeftPart := Data;
end;

class procedure TDnHttpParser.StripSpaces(var Data: String);
var I: Integer;
begin
//Detect first spaces
  for i := 1 to Length(Data) do
    if Data[i] <> '' then
      break;

  if I <= Length(Data) then
  begin
    if I <> 1 then
      Delete(Data, 1, I-1);
  end else
    Data := '';

//Detect last spaces
  for  I:= Length(Data) downto 1 do
    if Data[i] <> '' then
     break;

  if I > 0 then
  begin
    if I <> Length(Data) then
      Delete(Data, I, Length(Data) - I);
  end else
    Data := '';
end;

procedure GetLexem( Data: String; StartWith: Integer; var Lexem: String;
                    var FinishWith: Integer);
var i, len: Integer;
begin
  len := Length(Data);
  i := StartWith;
  while not (Data[i] in ['/', '\', ':', '?', '@']) and (i <= len) do
  begin
    Inc(i);
  end;
  Lexem := Copy(Data, StartWith, i-StartWith);
  FinishWith := i;
end;

procedure GetDelimiter( Data: String; StartWith: Integer; var Delimiter: String;
                        var FinishWith: Integer);
var i, len: Integer;
begin
  len := Length(Data);
  i := StartWith;
  while (Data[i] in ['/', '\', ':', '?', '@']) and (i <= len) do
  begin
    if (Data[i] = '@') and (i = StartWith) then
    begin
      FinishWith := i + 1;
      Delimiter := '@';
      Exit;
    end else
    if (Data[i] = '@') and (i <> StartWith) then
      break;
    Inc(i);
  end;
  Delimiter := Copy(Data, StartWith, i-StartWith);
  FinishWith := i;
end;

class function TDnHttpParser.IsAbsoluteUrl(const Url: String): Boolean;
begin
  Result := Pos('://', Url) <> 0;
end;

class procedure TDnHttpParser.ParseRelativeUrl(const Url: String; var Path, Query: String);
var QueryPos : Integer;
begin
  if Url = '*' then
  begin
    Path := '*';
    Query := '';
    Exit;
  end;
  
  QueryPos := Pos('?', Url);
  if QueryPos = 0 then
  begin
    Path := Url;
    Query := '';
  end else
  begin
    Path := Copy(Url, 1, QueryPos - 1);
    Query := Copy(Url, QueryPos, Length(Url) - QueryPos);
  end;
end;



class procedure TDnHttpParser.ParseAbsoluteUrl(const Url: String; var Protocol, User, Password,
                                        Host, Port, Path, Query: String;
                                        var UserExists, PasswordExists: Boolean);
var LastDelimiterPos, Fp: Integer;
    Lexem: String;
    Delimiter: String;
    //PossiblePassword,
    //PossibleHost: String;
    //PossiblePasswordExists: Boolean;
    QueryStart, PathStart: PChar;
    //URLLen: Integer;
begin
  Protocol := ''; User := ''; Password := '';
  Host := ''; Port := ''; Query := '';
  UserExists := False; PasswordExists := False;


  GetLexem(Url, 1, Lexem, Fp);
  LastDelimiterPos := Fp;
  GetDelimiter(Url, Fp, Delimiter, Fp); 
  if Delimiter = '://' then
  begin
    Protocol := LowerCase(Lexem);
    GetLexem(Url, Fp, Lexem, Fp); 
    LastDelimiterPos := Fp;
    GetDelimiter(Url, Fp, Delimiter, Fp); 
  end else
    Protocol := LowerCase('http');
  if Delimiter = '@' then
  begin
    UserExists := True;
    User := Lexem;

    GetLexem(Url, Fp, Lexem, Fp);
    LastDelimiterPos := Fp;
    GetDelimiter(Url, Fp, Delimiter, Fp);
    if Delimiter = ':' then
    begin
      Password := Lexem; PasswordExists := True;
      GetLexem(Url, Fp, Lexem, Fp);
      LastDelimiterPos := Fp;
      GetDelimiter(Url, Fp, Delimiter, Fp);
    end;
  end else
  if Delimiter = ':' then
  begin
    PasswordExists := True; Password := Lexem;
    GetLexem(Url, Fp, Lexem, Fp);
    LastDelimiterPos := Fp;
    GetDelimiter(Url, Fp, Delimiter, Fp);
  end;

  while Delimiter = '.' do
  begin
    Host := Host + Lexem + Delimiter;
    GetLexem(Url, Fp, Lexem, Fp);
    GetDelimiter(Url, Fp, Delimiter, Fp);
  end;
  Host := Host + Lexem;

  if Delimiter = ':' then
  begin //port
    GetLexem(Url, Fp, Lexem, Fp);
    LastDelimiterPos := Fp;
    GetDelimiter(Url, Fp, Delimiter, Fp);
    Port := Lexem;
  end;
  //LastDelimiterPos := Fp;
  QueryStart := StrScan(PChar(Url) + LastDelimiterPos-1, '?');
  PathStart := StrScan(PChar(Url) + LastDelimiterPos-1, '/');
  if QueryStart = Nil then
  begin
    //QueryExists := False;
    Query := '';
    if PathStart = Nil then
      Path := ''
    else
    begin
      Path := Copy(Url, PathStart - PChar(Url) + 1, Length(Url) - (PathStart - PChar(Url)));
    end;
  end else
  begin
    //QueryExists := True
    Query := Copy(Url, QueryStart - PChar(Url) + 1, Length(Url) - (QueryStart - PChar(Url)));
    if PathStart = Nil then
      Path := ''
    else begin
      Path := Copy(Url, PathStart - PChar(Url) + 1, QueryStart - PathStart);
    end;
  end;
end;

//-----------------------------------------------------------------------------

end.
